from ._core import _Global_Carbon_Budget_Release

GCB2022 = _Global_Carbon_Budget_Release(
    name="Global Carbon Budget 2022",
    version="1.0",
    doi="10.18160/gcp-2022",
    doi_article="10.5194/essd-14-4811-2022",
    published="2022-11-11",
    citation="Global Carbon Project. (2022). Supplemental data of Global Carbon Budget 2022 (Version 1.0) [Data set]. Global Carbon Project. https://doi.org/10.18160/gcp-2022",
    citation_article="Friedlingstein, P., O'Sullivan, M., Jones, M. W., Andrew, R. M., Gregor, L., Hauck, J., Le Quéré, C., Luijkx, I. T., Olsen, A., Peters, G. P., Peters, W., Pongratz, J., Schwingshackl, C., Sitch, S., Canadell, J. G., Ciais, P., Jackson, R. B., Alin, S. R., Alkama, R., Arneth, A., Arora, V. K., Bates, N. R., Becker, M., Bellouin, N., Bittig, H. C., Bopp, L., Chevallier, F., Chini, L. P., Cronin, M., Evans, W., Falk, S., Feely, R. A., Gasser, T., Gehlen, M., Gkritzalis, T., Gloege, L., Grassi, G., Gruber, N., Gürses, Ö., Harris, I., Hefner, M., Houghton, R. A., Hurtt, G. C., Iida, Y., Ilyina, T., Jain, A. K., Jersild, A., Kadono, K., Kato, E., Kennedy, D., Klein Goldewijk, K., Knauer, J., Korsbakken, J. I., Landschützer, P., Lefèvre, N., Lindsay, K., Liu, J., Liu, Z., Marland, G., Mayot, N., McGrath, M. J., Metzl, N., Monacci, N. M., Munro, D. R., Nakaoka, S.-I., Niwa, Y., O'Brien, K., Ono, T., Palmer, P. I., Pan, N., Pierrot, D., Pocock, K., Poulter, B., Resplandy, L., Robertson, E., Rödenbeck, C., Rodriguez, C., Rosan, T. M., Schwinger, J., Séférian, R., Shutler, J. D., Skjelvan, I., Steinhoff, T., Sun, Q., Sutton, A. J., Sweeney, C., Takao, S., Tanhua, T., Tans, P. P., Tian, X., Tian, H., Tilbrook, B., Tsujino, H., Tubiello, F., van der Werf, G. R., Walker, A. P., Wanninkhof, R., Whitehead, C., Willstrand Wranne, A., Wright, R., Yuan, W., Yue, C., Yue, X., Zaehle, S., Zeng, J., and Zheng, B.: Global Carbon Budget 2022, Earth Syst. Sci. Data, 14, 4811–4900, https://doi.org/10.5194/essd-14-4811-2022, 2022.",
    license="CC BY 4.0",
    filename="Global_Carbon_Budget_2022v1.0.xlsx",
    url="https://data.icos-cp.eu/licence_accept?ids=%5B%221umMtgeUlhS2Y1YW_Qp94bu3%22%5D",
    known_hash="d6e98cb607949614b6635616fd0a7de1bbb790fccaeaa96b9ff183db216c2b4d",
    sheets=[
        {
            "sheet_name": "Global Carbon Budget",
            "skiprows": 20,
        },
        {
            "sheet_name": "Historical Budget",
            "skiprows": 15,
        },
        {"sheet_name": "Fossil Emissions by Category", "skiprows": 8},
        {
            "sheet_name": "Land-Use Change Emissions",
            "skiprows": 26,
            "tables": [
                {
                    "table_name": "GCB",
                    "skiprows": 29,
                    "columns": "A:D",
                },
                {
                    "table_name": "H&N",
                    "skiprows": 29,
                    "columns": "A,F:H",
                },
                {
                    "table_name": "BLUE",
                    "skiprows": 29,
                    "columns": "A,I:K",
                },
                {
                    "table_name": "OSCAR",
                    "skiprows": 29,
                    "columns": "A,L:N",
                },
                {
                    "table_name": "Individual models",
                    "skiprows": list(range(28)) + [29],
                    "columns": "A,P:AE,AG:AH",
                },
            ],
        },
        {
            "sheet_name": "Ocean Sink",
            "skiprows": 28,
            "tables": [
                {"table_name": "GCB", "skiprows": 30, "columns": "A:C"},
                {
                    "table_name": "Individual models",
                    "skiprows": 30,
                    "columns": "A,E:N,P:Q",
                },
                {
                    "table_name": "Data-based products",
                    "skiprows": 30,
                    "columns": "A,S:AB",
                },
            ],
        },
        {
            "sheet_name": "Terrestrial Sink",
            "skiprows": 21,
            "tables": [
                {"table_name": "GCB", "skiprows": 23, "columns": "A:B"},
                {
                    "table_name": "Individual models",
                    "skiprows": 23,
                    "columns": "A,D:S,U:V",
                },
            ],
        },
        {"sheet_name": "Cement Carbonation Sink", "skiprows": 9, "columns": "A,B,D,E"},
    ],
)



from dataclasses import dataclass
from zipfile import ZipFile

import pandas as pd
import pooch


@dataclass
class _CedsRelease(dict):
    name: str
    citation: str
    doi: str
    published: str
    license: str
    entities: list[str]
    tables: list[dict]

    def __repr__(self):
        newline = "\n"

        return f"""{self.name}

License: {self.license}
https://doi.org/{self.doi}

Citation:
{self.citation}

{len(self.entities)} entities:
{newline.join([f'- "{k}"' for k in self.entities])}"""

    def __post_init__(self):
        for entity in self.entities:
            if entity not in self.keys():
                self[entity] = {}

            for table in self.tables:
                for key, path_pattern in table["patterns"].items():
                    self[entity][key] = _CedsTable(
                        entity=entity,
                        filename=table["zipfile"]["filename"],
                        url=table["zipfile"]["url"],
                        hash=table["zipfile"]["hash"],
                        path_pattern=(
                            path_pattern(entity)
                            if callable(path_pattern)
                            else path_pattern
                        ),
                    )


@dataclass
class _CedsTable:
    entity: str
    filename: str
    url: str
    hash: str
    path_pattern: str

    def __repr__(self):
        return f"""<Data from: {self.path_pattern.format(entity=self.entity)}>"""

    def _get_file_path(self):
        return pooch.retrieve(
            path=pooch.os_cache("openclimatedata/ceds"),
            fname=self.filename,
            url=self.url,
            known_hash=self.hash,
        )

    def _load_csv_from_zip(self, path_pattern):
        file_path = self._get_file_path()
        with ZipFile(file_path) as zip_file:
            dtype = {
                "country": "category",
                "iso": "category",
                "em": "category",
                "units": "category",
                "sector": "category",
                "fuel": "category",
            }

            path = self.path_pattern.format(entity=self.entity)
            # Workaround: in v_2025_03_18 NMVOC file is named as `emissions`, others `estimates`
            if path.endswith(
                "NMVOC_CEDS_estimates_by_country_CEDS_sector_fuel_v_2025_03_18.csv"
            ):
                path = path.replace("estimates", "emissions")
            df = pd.read_csv(
                zip_file.open(path),
                dtype=dtype,
            )

        return df

    def to_dataframe(self):
        return self._load_csv_from_zip(self.path_pattern)

    def to_long_dataframe(self):
        """
        Turn CEDS data into a long dataframe.
        """
        df = self.to_dataframe()

        id_vars = [c for c in df.columns if not c.startswith("X")]

        df.columns = [int(c[1:]) if c.startswith("X") else c for c in df.columns]

        df = df.melt(
            id_vars=id_vars,
            var_name="year",
            value_name="value",
        )
        df["year"] = df["year"].astype("int32")
        return df

    def to_ocd(self):
        """Return a long DataFrame with standardized codes and column names."""
        df = self.to_long_dataframe()
        column_names = {
            "sector": "category",
            "em": "entity",
            "units": "unit",
        }
        if "iso" in df.columns:
            column_names["iso"] = "code"
        elif "country" in df.columns:
            column_names["country"] = "code"
        df = df.rename(columns=column_names)

        if "code" in df.columns:
            df["code"] = df["code"].cat.rename_categories(
                {"global": "XIT", "srb (kosovo)": "XKX"}
            )
            df["code"] = df["code"].cat.rename_categories(str.upper)
        return df


CEDS = {
    "v_2025_03_18": _CedsRelease(
        name="CEDS v_2025_03_18 Aggregate Data",
        doi="10.5281/zenodo.15059443",
        published="2025-03-18",
        citation="""Hoesly, R., Smith, S. J., Ahsan, H., Prime, N., O'Rourke, P., Crippa, M., Klimont, Z., Guizzardi, D., Feng, L., Harkins, C., MCDONALD, B., & Wang, S. (2025). CEDS v_2025_03_18 Aggregate Data (v_2025_03_18) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.15059443""",
        license="CC BY 4.0",
        entities=[
            "BC",
            "CH4",
            "CO",
            "CO2",
            "N2O",
            "NH3",
            "NMVOC",
            "NOx",
            "OC",
            "SO2",
        ],
        tables=[
            {
                "zipfile": {
                    "filename": "CEDS_v_2025_03_18_aggregate.zip",
                    "url": "https://zenodo.org/records/15059443/files/CEDS_v_2025_03_18_aggregate.zip",
                    "hash": "md5:262b30b8f88708d6a567dbb58ec623a5",
                },
                "patterns": {
                    "by_country": "CEDS_v_2025_03_18_aggregate/{entity}_CEDS_estimates_by_country_v_2025_03_18.csv",
                    "by_sector": "CEDS_v_2025_03_18_aggregate/{entity}_CEDS_estimates_by_country_sector_v_2025_03_18.csv",
                    "global_by_sector": "CEDS_v_2025_03_18_aggregate/{entity}_CEDS_global_estimates_by_sector_v_2025_03_18.csv",
                    "global_by_sector_fuel": "CEDS_v_2025_03_18_aggregate/{entity}_CEDS_global_estimates_by_sector_fuel_v_2025_03_18.csv",
                },
            },
            {
                "zipfile": {
                    "filename": "CEDS_v_2025_03_18_detailed.zip",
                    "url": "https://zenodo.org/records/15059443/files/CEDS_v_2025_03_18_detailed.zip",
                    "hash": "md5:ee40a44dfca8b9c4751119c3d9daf22f",
                },
                "patterns": {
                    "by_sector_fuel": "CEDS_v_2025_03_18_detailed/{entity}_CEDS_estimates_by_country_CEDS_sector_fuel_v_2025_03_18.csv",
                },
            },
            {
                "zipfile": {
                    "filename": "CEDS_v_2025_03_18_supplementary_bunkers.zip",
                    "url": "https://zenodo.org/records/15059443/files/CEDS_v_2025_03_18_supplementary_bunkers.zip",
                    "hash": "md5:6fc40fdccb1482768887141de20a1da6",
                },
                "patterns": {
                    "bunkers": "CEDS_v_2025_03_18_supplementary_bunkers/S.{entity}_bunker_estimates_v_2025_03_18.csv",
                },
            },
            {
                "zipfile": {
                    "filename": "CEDS_v_2025_03_18_supplementary_extension.zip",
                    "url": "https://zenodo.org/records/15059443/files/CEDS_v_2025_03_18_supplementary_extension.zip",
                    "hash": "md5:5c2f99d3f18dbd896bb240f8a6c770ce",
                },
                "patterns": {
                    "extension_by_fuel": "CEDS_v_2025_03_18_supplementary_extension/{entity}_Extension_CEDS_global_estimates_by_fuel_v_2025_03_18.csv",
                    "extension_by_sector_fuel": "CEDS_v_2025_03_18_supplementary_extension/{entity}_Extension_CEDS_global_estimates_by_sector_fuel_v_2025_03_18.csv",
                    "extension_by_sector": "CEDS_v_2025_03_18_supplementary_extension/{entity}_Extension_CEDS_global_estimates_by_sector_v_2025_03_18.csv",
                },
            },
        ],
    ),
    "v_2024_07_08": _CedsRelease(
        name="CEDS v_2024_07_08 Release Emission Data",
        doi="10.5281/zenodo.12803197",
        published="2024-07-24",
        citation="""Hoesly, R., Smith, S. J., Prime, N., Ahsan, H., Suchyta, H., O'Rourke, P., Crippa, M., Klimont, Z., Guizzardi, D., Behrendt, J., Feng, L., Harkins, C., McDonald, B., Mott, A., McDuffie, A., Nicholson, M., & Wang, S. (2024). CEDS v_2024_07_08 Release Emission Data (v_2024_07_08) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.12803197""",
        license="CC BY 4.0",
        entities=[
            "BC",
            "CH4",
            "CO",
            "CO2",
            "N2O",
            "NH3",
            "NMVOC",
            "NOx",
            "OC",
            "SO2",
        ],
        tables=[
            {
                "zipfile": {
                    "filename": "CEDS_v_2024_07_08_aggregate.zip",
                    "url": "https://zenodo.org/records/12803197/files/CEDS_v_2024_07_08_aggregate.zip",
                    "hash": "md5:4ad029d409b8a66ed6f3d85319c66964",
                },
                "patterns": {
                    "by_country": "{entity}_CEDS_emissions_by_country_v2024_07_08.csv",
                    "by_sector": "{entity}_CEDS_emissions_by_country_sector_v2024_07_08.csv",
                    "global_by_sector": "{entity}_CEDS_global_emissions_by_sector_v2024_07_08.csv",
                    "global_by_sector_fuel": "{entity}_CEDS_global_emissions_by_sector_fuel_v2024_07_08.csv",
                },
            },
            {
                "zipfile": {
                    "filename": "CEDS_v_2024_07_08_detailed.zip",
                    "url": "https://zenodo.org/records/12803197/files/CEDS_v_2024_07_08_detailed.zip",
                    "hash": "md5:558811c1a1c8e6997dcb266c18f9c093",
                },
                "patterns": {
                    "by_sector_fuel": "{entity}_CEDS_emissions_by_country_CEDS_sector_fuel_v2024_07_08.csv",
                },
            },
            {
                "zipfile": {
                    "filename": "CEDS_v_2024_07_08_supplementary_bunkers.zip",
                    "url": "https://zenodo.org/records/12803197/files/CEDS_v_2024_07_08_supplementary_bunkers.zip",
                    "hash": "md5:8b90232036c5aea8ba1376c7450d23a4",
                },
                "patterns": {
                    "bunkers": "S.{entity}_bunker_emissions_v2024_07_08.csv",
                },
            },
        ],
    ),
    "v_2024_04_01": _CedsRelease(
        name="CEDS v_2024_04_01 Release Emission Data",
        doi="10.5281/zenodo.10904361",
        published="2024-04-01",
        citation="""Hoesly, R., & Smith, S. (2024). CEDS v_2024_04_01 Release Emission Data (v_2024_04_01) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.10904361""",
        license="CC BY 4.0",
        entities=[
            "BC",
            "CH4",
            "CO",
            "CO2",
            "N2O",
            "NH3",
            "NMVOC",
            "NOx",
            "OC",
            "SO2",
        ],
        tables=[
            {
                "zipfile": {
                    "filename": "CEDS_v_2024_04_01_aggregate.zip",
                    "url": "https://zenodo.org/records/10904361/files/CEDS_v_2024_04_01_aggregate.zip",
                    "hash": "md5:636752881e915244bb96b4793d9fb121",
                },
                "patterns": {
                    "by_country": "CEDS_v_2024_04_01_aggregate/{entity}_CEDS_emissions_by_country_v2024_04_01.csv",
                    "by_sector": "CEDS_v_2024_04_01_aggregate/{entity}_CEDS_emissions_by_country_sector_v2024_04_01.csv",
                    "global_by_sector": "CEDS_v_2024_04_01_aggregate/{entity}_CEDS_global_emissions_by_sector_v2024_04_01.csv",
                    "global_by_sector_fuel": "CEDS_v_2024_04_01_aggregate/{entity}_CEDS_global_emissions_by_sector_fuel_v2024_04_01.csv",
                },
            },
            {
                "zipfile": {
                    "filename": "CEDS_v_2024_04_01_detailed.zip",
                    "url": "https://zenodo.org/records/10904361/files/CEDS_v_2024_04_01_detailed.zip",
                    "hash": "md5:8310c3c9ad0da67c11f61168ca7d265d",
                },
                "patterns": {
                    "by_sector_fuel": "CEDS_v_2024_04_01_detailed/{entity}_CEDS_emissions_by_country_CEDS_sector_fuel_v2024_04_01.csv",
                },
            },
            {
                "zipfile": {
                    "filename": "CEDS_v_2024_04_01_supplementary_bunkers.zip",
                    "url": "https://zenodo.org/records/10904361/files/CEDS_v_2024_04_01_supplementary_bunkers.zip",
                    "hash": "md5:49d6bbe523409ca61cc84854fe4a27b2",
                },
                "patterns": {
                    "bunkers": "CEDS_v_2024_04_01_supplementary_bunkers/S.{entity}_bunker_emissions_v2024_04_01.csv",
                },
            },
        ],
    ),
    "v_2021_04_21": _CedsRelease(
        name="CEDS v_2021_04_21 Release Emission Data",
        doi="10.5281/zenodo.4741285",
        published="2021-04-06",
        # TODO fix the citation with the correct version number? see https://github.com/JGCRI/CEDS/issues/48
        citation="""O'Rourke, P. R., Smith, S. J., Mott, A., Ahsan, H., McDuffie, E. E., Crippa, M., Klimont, Z., McDonald, B., Wang, S., Nicholson, M. B., Feng, L., & Hoesly, R. M. (2021). CEDS v_2021_04_21 Release Emission Data (v_2021_02_05) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.4741285""",
        license="CC BY 4.0",
        entities=[
            "BC",
            "CH4",
            "CO",
            "CO2",
            "N2O",
            "NH3",
            "NMVOC",
            "NOx",
            "OC",
            "SO2",
        ],
        tables=[
            {
                "zipfile": {
                    "filename": "CEDS_v2021-04-21_emissions.zip",
                    "url": "https://zenodo.org/records/4741285/files/CEDS_v2021-04-21_emissions.zip",
                    "hash": "md5:01659c651754a66ddf3d79715a2ba841",
                },
                "patterns": {
                    "by_country": "{entity}_CEDS_emissions_by_country_2021_04_21.csv",
                    "by_sector": "{entity}_CEDS_emissions_by_sector_country_2021_04_21.csv",
                    "by_fuel": "{entity}_CEDS_emissions_by_country_fuel_2021_04_21.csv",
                },
            },
        ],
    ),
    "v_2021_02_05": _CedsRelease(
        name="CEDS v_2021_02_05 Release Emission Data",
        doi="10.5281/zenodo.4509372",
        published="2021-02-05",
        citation="""O'Rourke, P. R., Smith, S. J., Mott, A., Ahsan, H., McDuffie, E. E., Crippa, M., Klimont, Z., McDonald, B., Wang, S., Nicholson, M. B., Feng, L., & Hoesly, R. M. (2021). CEDS v_2021_02_05 Release Emission Data (v_2021_02_05) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.4509372""",
        license="CC BY 4.0",
        entities=[
            "BC",
            "CH4",
            "CO",
            "CO2",
            "N2O",
            "NH3",
            "NMVOC",
            "NOx",
            "OC",
            "SO2",
        ],
        tables=[
            {
                "zipfile": {
                    "filename": "CEDS_v2021-02-05_emissions.zip",
                    "url": "https://zenodo.org/records/4509372/files/CEDS_v2021-02-05_emissions.zip",
                    "hash": "md5:7054c1e1ca510015a37d6c1bb5934c9b",
                },
                "patterns": {
                    "by_country": "CEDS_v2021-02-05_emissions/{entity}_CEDS_emissions_by_country_2021_02_05.csv",
                    "by_sector": "CEDS_v2021-02-05_emissions/{entity}_CEDS_emissions_by_sector_country_2021_02_05.csv",
                    "by_fuel": "CEDS_v2021-02-05_emissions/{entity}_CEDS_emissions_by_country_fuel_2021_02_05.csv",
                },
            },
        ],
    ),
    "v_2020_09_11": _CedsRelease(
        name="CEDS v_2020_09_11 Pre-Release Emission Data",
        doi="10.5281/zenodo.4025316",
        published="2020-09-11",
        citation="""O'Rourke, P. R., Smith, S. J., McDuffie, E. E., Klimont, Z., Crippa, M., Mott, A., Wang, S., Nicholson, M. B., Feng, L., & Hoesly, R. M. (2020). CEDS v_2020_09_11 Pre-Release Emission Data (v_2020_09_11) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.4025316""",
        license="CC BY 4.0",
        entities=["BC", "CO", "NH3", "NMVOC", "NOx", "OC", "SO2"],
        tables=[
            {
                "zipfile": {
                    "filename": "CEDS_v_2020_09_11_emissions.zip",
                    "url": "https://zenodo.org/records/4025316/files/CEDS_v_2020_09_11_emissions.zip",
                    "hash": "md5:0c7f4bfc5eafcd7510920fd0b8bbdd16",
                },
                "patterns": {
                    "by_country": "{entity}_CEDS_emissions_by_country_2020_09_11.csv",
                    "by_sector": "{entity}_CEDS_emissions_by_sector_country_2020_09_11.csv",
                    "by_fuel": "{entity}_CEDS_emissions_by_country_fuel_2020_09_11.csv",
                },
            },
        ],
    ),
    "v_2019_12_23": _CedsRelease(
        name="CEDS v_2019_12_23 Emission Data",
        doi="10.5281/zenodo.3606753",
        published="2020-01-13",
        citation="""Hoesly, R. M., O'Rourke, P. R., Smith, S. J., Feng, L., Klimont, Z., Janssens-Maenhout, G., Pitkanen, T., Seibert, J. J., Vu, L., Andres, R. J., Bolt, R. M., Bond, T. C., Dawidowski, L., Kholod, N., Kurokawa, J.-. ichi ., Li, M., Liu, L., Lu, Z., Moura, M. C. P., Zhang, Q., Goldstein, B., Muwan, P. (2020). CEDS v_2019_12_23 Emission Data (v_2019_12_23) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.3606753""",
        license="CC BY 4.0",
        entities=["BC", "CH4", "CO", "CO2", "NH3", "NMVOC", "NOx", "OC", "SO2"],
        tables=[
            {
                "zipfile": {
                    "filename": "CEDS_v_2019_12_23-final_emissions.zip",
                    "url": "https://zenodo.org/records/3606753/files/CEDS_v_2019_12_23-final_emissions.zip",
                    "hash": "md5:830ac6fbc5ba24885acecf1aa6567db8",
                },
                "patterns": {
                    "by_country": "CEDS_v_2019_12_23-final_emissions/CEDS_{entity}_emissions_by_country_v_2019_12_23.csv",
                    "by_sector": "CEDS_v_2019_12_23-final_emissions/CEDS_{entity}_emissions_by_country_CEDS_sector_v_2019_12_23.csv",
                    "global_by_fuel": "CEDS_v_2019_12_23-final_emissions/CEDS_{entity}_global_emissions_by_fuel_v_2019_12_23.csv",
                },
            },
        ],
    ),
    # "As of 2 August 2018, a corrected supplement is available for download,
    # which now includes missing data files, most notably, CO2 and CH4 emissions." (Corrected Supplement Explanation.docx)
    # The data supplementary zip file contains file with filenames starting with 2017_05_18 and 2016_07_26
    # The CH4_Extension files (by region) are not included below
    "v_2018_08_02": _CedsRelease(
        name="CEDS Corrected Supplemental Data",
        doi="10.5194/gmd-11-369-2018",
        published="2018-08-02",
        citation="""Hoesly, R. M., Smith, S. J., Feng, L., Klimont, Z., Janssens-Maenhout, G., Pitkanen, T., Seibert, J. J., Vu, L., Andres, R. J., Bolt, R. M., Bond, T. C., Dawidowski, L., Kholod, N., Kurokawa, J.-I., Li, M., Liu, L., Lu, Z., Moura, M. C. P., O'Rourke, P. R., and Zhang, Q.: Historical (1750–2014) anthropogenic emissions of reactive gases and aerosols from the Community Emissions Data System (CEDS), Geosci. Model Dev., 11, 369-408, https://doi.org/10.5194/gmd-11-369-2018, 2018.""",
        license="CC BY 3.0",
        entities=["SO2", "NOx", "BC", "OC", "NH3", "NMVOC", "CO", "CO2", "CH4"],
        tables=[
            {
                "zipfile": {
                    "filename": "gmd-11-369-2018-supplement.zip",
                    "url": "https://doi.org/10.5194/gmd-11-369-2018-supplement",
                    "hash": "md5:af7e686222c2c98cd3a38f2b9a5dbd24",
                },
                "patterns": {
                    "by_country": lambda entity: (
                        "Supplemental_Data_Correction/Data Supplement/{entity}_CEDS_emissions_by_country_v2016_07_26.csv"
                        if entity not in ["CO2", "CH4"]
                        else "Supplemental_Data_Correction/Data Supplement/{entity}_CEDS_emissions_by_country_v2017_05_18.csv"
                    ),
                    "by_sector": lambda entity: (
                        "Supplemental_Data_Correction/Data Supplement/{entity}_CEDS_emissions_by_sector_country_v2016_07_26.csv"
                        if entity not in ["CO2", "CH4"]
                        else "Supplemental_Data_Correction/Data Supplement/{entity}_CEDS_emissions_by_sector_country_v2017_05_18.csv"
                    ),
                    "global_by_sector": lambda entity: (
                        "Supplemental_Data_Correction/Data Supplement/{entity}_CEDS_emissions_by_sector_v2016_07_26.csv"
                        if entity not in ["CO2", "CH4"]
                        else "Supplemental_Data_Correction/Data Supplement/{entity}_CEDS_emissions_by_sector_v2017_05_18.csv"
                    ),
                },
            }
        ],
    ),
}



from dataclasses import dataclass
from typing import TypedDict

import pandas as pd
import pooch

newline = "\n"


class _File(TypedDict):
    filename: str
    known_hash: str
    note: str


class _PRIMAPHIST_2_Release(dict):
    def __init__(
        self,
        name: str,
        version: str,
        published: str,
        doi: str,
        citation: str,
        doi_article: str,
        citation_article: str,
        files: dict[str, _File],
        license: str,
    ):
        self.name = name
        self.version = version
        self.published = published
        self.doi = doi
        self.citation = citation
        self.doi_article = doi_article
        self.citation_article = citation_article
        self.files = files
        self.license = license

        for key, datafile in files.items():
            self[key] = PRIMAPHIST_2_File(
                filename=datafile["filename"],
                known_hash=datafile["known_hash"],
                note=datafile["note"],
                release=self,
            )

    def __repr__(self):
        return f"""{self.name}

License: {self.license}

https://doi.org/{self.doi}

Recommended citation:

{self.citation_article}

{self.citation}

Files:

{newline.join([f"{k} - {f['filename']}" for k, f in self.files.items()])}
        """


@dataclass
class PRIMAPHIST_2_File:
    filename: str
    known_hash: str
    note: str
    release: object

    def __repr__(self):
        return f"""{self.release.name}

License: {self.release.license}

https://doi.org/{self.release.doi}

Recommended citation:

{self.release.citation_article}

{self.release.citation}

File: {self.filename}

{self.note}"""

    def to_dataframe(self):
        full_path = pooch.retrieve(
            path=pooch.os_cache("openclimatedata/primap-hist"),
            url=f"doi:{self.release.doi}/{self.filename}",
            known_hash=self.known_hash,
            progressbar=True,
        )
        if self.release.version >= "2.3":
            dtypes = {
                "source": "category",
                "scenario (PRIMAP-hist)": "category",
                "provenance": "category",
                "area (ISO3)": "category",
                "entity": "category",
                "unit": "category",
                "category (IPCC2006_PRIMAP)": "category",
            }
        else:
            dtypes = {
                "source": "category",
                "scenario": "category",
                "provenance": "category",
                "area": "category",
                "entity": "category",
                "unit": "category",
                "category": "category",
            }
        return pd.read_csv(full_path, dtype=dtypes)

    def to_long_dataframe(self):
        df = self.to_dataframe()

        # Pre 2.5 provenance not included.
        if self.release.version >= "2.5":
            id_vars = [
                "source",
                "scenario (PRIMAP-hist)",
                "provenance",
                "area (ISO3)",
                "entity",
                "unit",
                "category (IPCC2006_PRIMAP)",
            ]
        # Changed column names in 2.3
        elif self.release.version >= "2.3":
            id_vars = [
                "source",
                "scenario (PRIMAP-hist)",
                "area (ISO3)",
                "entity",
                "unit",
                "category (IPCC2006_PRIMAP)",
            ]
        else:
            id_vars = [
                "scenario",
                "country",
                "category",
                "entity",
                "unit",
            ]

        df = df.melt(
            id_vars=id_vars,
            var_name="year",
            value_name="value",
        )
        df.year = df.year.astype("int32")
        return df

    def to_ocd(self):
        """Long DataFrame with all column names shortened."""
        df = self.to_long_dataframe()
        if self.release.version >= "2.3":
            df = df.rename(
                columns={
                    "scenario (PRIMAP-hist)": "scenario",
                    "area (ISO3)": "code",
                    "category (IPCC2006_PRIMAP)": "category",
                }
            )
        else:
            df = df.rename(columns={"country": "code"})
        if "provenance" in df.columns:
            df = df.drop("provenance", axis=1)
        if self.release.version <= "2.2":
            df.category = df.category.cat.rename_categories(
                {
                    "IPCM0EL": "M.0.EL",
                    "IPCMAG": "M.AG",
                    "IPCMAGELV": "M.AG.ELV",
                    "IPC1": "1",
                    "IPC1A": "1.A",
                    "IPC1B": "1.B",
                    "IPC1B1": "1.B.1",
                    "IPC1B2": "1.B.2",
                    "IPC1B3": "1.B.3",
                    "IPC1C": "1.C",
                    "IPC2": "2",
                    "IPC2A": "2.A",
                    "IPC2B": "2.B",
                    "IPC2C": "2.C",
                    "IPC2D": "2.D",
                    "IPC2E": "2.E",
                    "IPC2F": "2.F",
                    "IPC2G": "2.G",
                    "IPC2H": "2.H",
                    "IPC3A": "3.A",
                    "IPC4": "4",
                    "IPC5": "5",
                }
            )

            # Use entity style of >=2.3
            df.entity = df.entity.cat.rename_categories(
                {
                    "FGASES": "FGASES (SARGWP100)",
                    "FGASESAR4": "FGASES (AR4GWP100)",
                    "HFCS": "HFCS (SARGWP100)",
                    "HFCSAR4": "HFCS (AR4GWP100)",
                    "PFCS": "PFCS (SARGWP100)",
                    "PFCSAR4": "PFCS (AR4GWP100)",
                    "KYOTOGHG": "KYOTOGHG (SARGWP100)",
                    "KYOTOGHGAR4": "KYOTOGHG (AR4GWP100)",
                }
            )
        return df



import math
import os

import pytest
from pytest import approx

from openclimatedata import GCB_Fossil_Emissions

GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"


versions = GCB_Fossil_Emissions.keys()


def test_gcb_fossil():
    for version in versions:
        assert GCB_Fossil_Emissions[version].name
        assert GCB_Fossil_Emissions[version].doi
        assert GCB_Fossil_Emissions[version].published


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_gcb_fossil_2023v28():
    df = GCB_Fossil_Emissions["2023v28"].to_dataframe()
    assert df.iloc[0]["Total"] == 0
    assert df.iloc[-1]["Per Capita"] == approx(4.663492)

    ocdf = GCB_Fossil_Emissions["2023v28"].to_ocd()
    # First and last value should be the same after re-shaping.
    assert ocdf.iloc[0]["value"] == df.iloc[0]["Total"]
    assert ocdf.iloc[-1]["value"] == df.iloc[-1]["Per Capita"]


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_gcb_fossil_2023v36():
    df = GCB_Fossil_Emissions["2023v36"].to_dataframe()
    assert df.iloc[0]["Total"] == 0
    assert df.iloc[-1]["Per Capita"] == approx(4.658219)

    ocdf = GCB_Fossil_Emissions["2023v36"].to_ocd()
    # First and last value should be the same after re-shaping.
    assert ocdf.iloc[0]["value"] == df.iloc[0]["Total"]
    assert ocdf.iloc[-1]["value"] == df.iloc[-1]["Per Capita"]


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_gcb_fossil_all_codes_set():
    df = GCB_Fossil_Emissions["2023v36"].to_dataframe()
    leeward_islands = df[df.Country == "Leeward Islands"]

    df_long = GCB_Fossil_Emissions["2023v36"].to_long_dataframe()

    assert (
        df_long[
            (df_long.Code == "Leeward Islands")
            & (df_long.Year == 1950)
            & (df_long.Category == "Total")
        ].Value.values[0]
        == leeward_islands[leeward_islands.Year == 1950].Total.values[0]
    )

    assert sum(df_long.Code.isnull()) == 0


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_gcb_fossil_2023v43():
    df = GCB_Fossil_Emissions["2023v43"].to_dataframe()
    # This version has Cement with zero and Total as `nan`.
    assert math.isnan(df.iloc[0]["Total"])
    assert df.iloc[-1]["Per Capita"] == approx(4.658365)

    ocdf = GCB_Fossil_Emissions["2023v43"].to_ocd()
    # First and last value should be the same after re-shaping.
    assert math.isnan(ocdf.iloc[0]["value"]) and ocdf.iloc[0]["category"] == "Total"
    assert ocdf.iloc[-1]["value"] == df.iloc[-1]["Per Capita"]


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_gcb_fossil_2024v17():
    df = GCB_Fossil_Emissions["2024v17"].to_dataframe()
    # This version has Cement with zero and Total as `nan`.
    assert math.isnan(df.iloc[0]["Total"])
    assert df.iloc[-1]["Per Capita"] == approx(4.697341)

    ocdf = GCB_Fossil_Emissions["2024v17"].to_ocd()
    # First and last value should be the same after re-shaping.
    assert math.isnan(ocdf.iloc[0]["value"]) and ocdf.iloc[0]["category"] == "Total"
    assert ocdf.iloc[-1]["value"] == df.iloc[-1]["Per Capita"]


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_gcb_fossil_2024v18():
    df = GCB_Fossil_Emissions["2024v18"].to_dataframe()
    # This version has Cement with zero and Total as `nan`.
    assert math.isnan(df.iloc[0]["Total"])
    assert df.iloc[-1]["Per Capita"] == approx(4.697341)

    ocdf = GCB_Fossil_Emissions["2024v18"].to_ocd()
    # First and last value should be the same after re-shaping.
    assert math.isnan(ocdf.iloc[0]["value"]) and ocdf.iloc[0]["category"] == "Total"
    assert ocdf.iloc[-1]["value"] == df.iloc[-1]["Per Capita"]


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_unique_code_name_combinations():
    # Up to version 2024v17
    # Saint Kitts and Nevis is included twice in GCB Fossil with code 'KNA'.
    # The second time the name is 'St-Kitts-Nevis-Anguilla'.
    # This should be handled in the long DataFrame which drops the country.
    for version in versions:
        df = GCB_Fossil_Emissions[version].to_dataframe()
        dfl = GCB_Fossil_Emissions[version].to_long_dataframe()
        assert len(df.Country.unique()) == len(dfl.Code.unique())


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_kosovo_country_code():
    df = GCB_Fossil_Emissions["2023v43"].to_long_dataframe()

    assert len(df[df.Code == "XKX"]) == 0
    assert len(df[df.Code == "KSV"]) > 0

    ocdf = GCB_Fossil_Emissions["2023v43"].to_ocd()
    assert len(ocdf[ocdf.code == "KSV"]) == 0
    assert len(ocdf[ocdf.code == "XKX"]) > 0

@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_gcb_fossil_2025v15():
    df = GCB_Fossil_Emissions["2025v15"].to_dataframe()
    assert df.iloc[-1]["Per Capita"] == approx(4.729075)

    ocdf = GCB_Fossil_Emissions["2025v15"].to_ocd()
    assert ocdf.iloc[-1]["value"] == df.iloc[-1]["Per Capita"]



from ._core import _Global_Carbon_Budget_Release

GCB2021 = _Global_Carbon_Budget_Release(
    name="Global Carbon Budget 2021",
    version="1.0",
    doi="10.18160/gcp-2021",
    doi_article="10.5194/essd-14-1917-2022",
    published="2022-03-21",
    citation="Global Carbon Project. (2021). Supplemental data of Global Carbon Budget 2021 (Version 1.0) [Data set]. Global Carbon Project. https://doi.org/10.18160/gcp-2021",
    citation_article="Friedlingstein, P., Jones, M. W., O'Sullivan, M., Andrew, R. M., Bakker, D. C. E., Hauck, J., Le Quéré, C., Peters, G. P., Peters, W., Pongratz, J., Sitch, S., Canadell, J. G., Ciais, P., Jackson, R. B., Alin, S. R., Anthoni, P., Bates, N. R., Becker, M., Bellouin, N., Bopp, L., Chau, T. T. T., Chevallier, F., Chini, L. P., Cronin, M., Currie, K. I., Decharme, B., Djeutchouang, L. M., Dou, X., Evans, W., Feely, R. A., Feng, L., Gasser, T., Gilfillan, D., Gkritzalis, T., Grassi, G., Gregor, L., Gruber, N., Gürses, Ö., Harris, I., Houghton, R. A., Hurtt, G. C., Iida, Y., Ilyina, T., Luijkx, I. T., Jain, A., Jones, S. D., Kato, E., Kennedy, D., Klein Goldewijk, K., Knauer, J., Korsbakken, J. I., Körtzinger, A., Landschützer, P., Lauvset, S. K., Lefèvre, N., Lienert, S., Liu, J., Marland, G., McGuire, P. C., Melton, J. R., Munro, D. R., Nabel, J. E. M. S., Nakaoka, S.-I., Niwa, Y., Ono, T., Pierrot, D., Poulter, B., Rehder, G., Resplandy, L., Robertson, E., Rödenbeck, C., Rosan, T. M., Schwinger, J., Schwingshackl, C., Séférian, R., Sutton, A. J., Sweeney, C., Tanhua, T., Tans, P. P., Tian, H., Tilbrook, B., Tubiello, F., van der Werf, G. R., Vuichard, N., Wada, C., Wanninkhof, R., Watson, A. J., Willis, D., Wiltshire, A. J., Yuan, W., Yue, C., Yue, X., Zaehle, S., and Zeng, J.: Global Carbon Budget 2021, Earth Syst. Sci. Data, 14, 1917–2005, https://doi.org/10.5194/essd-14-1917-2022, 2022.",
    license="CC BY 4.0",
    filename="Global_Carbon_Budget_2021v1.0.xlsx",
    url="https://data.icos-cp.eu/licence_accept?ids=%5B%22Ayyw1HeihXdTUoO000dGcxrP%22%5D",
    known_hash="032cb0d477a28577535283b4d34746731acf19303eac44824564e492456a692a",
    sheets=[
        {
            "sheet_name": "Global Carbon Budget",
            "skiprows": 20,
        },
        {
            "sheet_name": "Historical Budget",
            "skiprows": 15,
        },
        {"sheet_name": "Fossil Emissions by Category", "skiprows": 8},
        {
            "sheet_name": "Land-Use Change Emissions",
            "skiprows": 27,
            "tables": [
                {
                    "table_name": "GCB",
                    "skiprows": 30,
                    "columns": "A:D",
                },
                {
                    "table_name": "H&N",
                    "skiprows": 30,
                    "columns": "A,F:H",
                },
                {
                    "table_name": "BLUE",
                    "skiprows": 30,
                    "columns": "A,I:K",
                },
                {
                    "table_name": "OSCAR",
                    "skiprows": 30,
                    "columns": "A,L:N",
                },
                {
                    "table_name": "Individual models",
                    "skiprows": list(range(29)) + [30],
                    "columns": "A,P:AF,AH:AI",
                },
            ],
        },
        {
            "sheet_name": "Ocean Sink",
            "skiprows": 25,
            "tables": [
                {"table_name": "GCB", "skiprows": 27, "columns": "A:C"},
                {
                    "table_name": "Individual models",
                    "skiprows": 27,
                    "columns": "A,E:L,N:O",
                },
                {
                    "table_name": "Data-based products",
                    "skiprows": 27,
                    "columns": "A,Q:Z",
                },
            ],
        },
        {
            "sheet_name": "Terrestrial Sink",
            "skiprows": 22,
            "tables": [
                {"table_name": "GCB", "skiprows": 24, "columns": "A:B"},
                {
                    "table_name": "Individual models",
                    "skiprows": 24,
                    "columns": "A,D:T,V:W",
                },
            ],
        },
        {"sheet_name": "Cement Carbonation Sink", "skiprows": 9, "columns": "A,B,D,E"},
    ],
)



import os

import pytest
from pytest import approx

from openclimatedata import PRIMAPHIST

GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"

versions = PRIMAPHIST.keys()


def test_primaphist():
    for version in versions:
        assert PRIMAPHIST[version].name
        assert PRIMAPHIST[version].doi
        assert PRIMAPHIST[version].published


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_primaphist_2_6_1():
    df = PRIMAPHIST["2.6.1"]["main"].to_dataframe()
    assert df.iloc[0]["1750"] == approx(0.00564)
    assert df.iloc[-1]["2023"] == approx(0.0)

    assert df.iloc[0]["provenance"] == "derived"

    ocdf = PRIMAPHIST["2.6.1"]["main"].to_ocd()
    # First and last value should be the same after re-shaping.
    assert ocdf.iloc[0]["value"] == df.iloc[0]["1750"]
    assert ocdf.iloc[-1]["value"] == df.iloc[-1]["2023"]


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_primaphist_2_6():
    df = PRIMAPHIST["2.6"]["main"].to_dataframe()
    assert df.iloc[0]["1750"] == approx(0.00564)
    assert df.iloc[-1]["2023"] == approx(0.0)

    assert df.iloc[0]["provenance"] == "derived"

    ocdf = PRIMAPHIST["2.6"]["main"].to_ocd()
    # First and last value should be the same after re-shaping.
    assert ocdf.iloc[0]["value"] == df.iloc[0]["1750"]
    assert ocdf.iloc[-1]["value"] == df.iloc[-1]["2023"]


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_primaphist_2_5_1():
    df = PRIMAPHIST["2.5.1"]["main"].to_dataframe()
    assert df.iloc[0]["1750"] == approx(0.00564)
    assert df.iloc[-1]["2022"] == approx(0.0)

    assert df.iloc[0]["provenance"] == "derived"

    ocdf = PRIMAPHIST["2.5.1"]["main"].to_ocd()
    # First and last value should be the same after re-shaping.
    assert ocdf.iloc[0]["value"] == df.iloc[0]["1750"]
    assert ocdf.iloc[-1]["value"] == df.iloc[-1]["2022"]


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_primaphist_2_5():
    df = PRIMAPHIST["2.5"]["main"].to_dataframe()
    assert df.iloc[0]["1750"] == approx(0.00564)
    assert df.iloc[-1]["2022"] == approx(0.0)

    assert df.iloc[0]["provenance"] == "measured"

    ocdf = PRIMAPHIST["2.5"]["main"].to_ocd()
    # First and last value should be the same after re-shaping.
    assert ocdf.iloc[0]["value"] == df.iloc[0]["1750"]
    assert ocdf.iloc[-1]["value"] == df.iloc[-1]["2022"]


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_primaphist_2_4_2():
    df = PRIMAPHIST["2.4.2"]["main"].to_dataframe()
    assert df.iloc[0]["1750"] == approx(0.00564)
    assert df.iloc[-1]["2021"] == approx(13.9)

    ocdf = PRIMAPHIST["2.4.2"]["main"].to_ocd()
    # First and last value should be the same after re-shaping.
    assert ocdf.iloc[0]["value"] == df.iloc[0]["1750"]
    assert ocdf.iloc[-1]["value"] == df.iloc[-1]["2021"]


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_primaphist_2_3():
    df = PRIMAPHIST["2.3"]["main"].to_dataframe()
    assert df.iloc[0]["1750"] == approx(0.00564)
    assert df.iloc[-1]["2019"] == approx(15.6)

    ocdf = PRIMAPHIST["2.3"]["main"].to_ocd()
    # First and last value should be the same after re-shaping.
    assert ocdf.iloc[0]["value"] == df.iloc[0]["1750"]
    assert ocdf.iloc[-1]["value"] == df.iloc[-1]["2019"]


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_primaphist_2_2():
    df = PRIMAPHIST["2.2"]["main"].to_dataframe()
    assert df.iloc[0]["1850"] == approx(1.53e-04)
    assert df.iloc[-1]["2018"] == approx(1.99e-02)

    ocdf = PRIMAPHIST["2.2"]["main"].to_ocd()
    # First and last value should be the same after re-shaping.
    assert ocdf.iloc[0]["value"] == df.iloc[0]["1850"]
    assert ocdf.iloc[-1]["value"] == df.iloc[-1]["2018"]

    ocdf_2_3 = PRIMAPHIST["2.3"]["main"].to_ocd()

    # 2.2. ocd table should use the category style of later versions, without 'IPC' prefix
    assert set(ocdf.category.unique()) == set(ocdf_2_3.category.unique())

    # Entities shold use the style of >=2.3 with GWP in parentheses
    entities = set(ocdf.entity.unique())
    assert set(
        [
            "KYOTOGHG (AR4GWP100)",
            "KYOTOGHG (SARGWP100)",
            "FGASES (AR4GWP100)",
            "FGASES (SARGWP100)",
            "HFCS (AR4GWP100)",
            "HFCS (SARGWP100)",
            "PFCS (AR4GWP100)",
            "PFCS (SARGWP100)",
        ]
    ).issubset(entities)

    assert not set(
        [
            "KYOTOGHG",
            "KYOTOGHGAR4",
            "FGASES",
            "FGASESAR4",
            "HFCS",
            "HFCSAR4",
            "PFCS",
            "PFCSAR4",
        ]
    ).issubset(entities)



from ._ceds import CEDS
from ._gcb_fossil import GCB_Fossil_Emissions
from ._gcb_national import GCB_National_Emissions
from ._global_carbon_budget import Global_Carbon_Budget
from ._primap_hist import PRIMAPHIST

__all__ = [
    "PRIMAPHIST",
    "CEDS",
    "Global_Carbon_Budget",
    "GCB_National_Emissions",
    "GCB_Fossil_Emissions",
]

from importlib.metadata import version

__version__ = version("openclimatedata")



from dataclasses import dataclass

import numpy as np
import pandas as pd
import pooch


@dataclass
class _GCB_Fossil:
    name: str
    citation: str
    doi: str
    published: str
    filename: str
    url: str
    hash: str
    filename_sources: str
    url_sources: str
    hash_sources: str
    license: str

    def __repr__(self):
        return f"""{self.name}
'{self.filename}'

License: {self.license}
https://doi.org/{self.doi}

{self.citation}"""

    def to_dataframe(self):
        file_path = pooch.retrieve(
            path=pooch.os_cache("openclimatedata/gcb-fossil"),
            fname=self.filename,
            url=self.url,
            known_hash=self.hash,
        )
        return pd.read_csv(
            file_path,
            encoding="latin-1",
            dtype={
                "Country": "category",
                "ISO 3166-1 alpha-3": "category",
                "UN M49": "category",
                "Year": "int32",
            },
        )

    def to_long_dataframe(self):
        """
        Turn GCB Fossil data into a long dataframe and add source metadata
        as a column. `ISO 3166-1 alpha-3` is renamed to `Code`.
        """
        df = self.to_dataframe()

        if "UN M49" in df.columns:
            df = df.drop("UN M49", axis=1)

        df["Country"] = df["Country"].astype("string")
        df["ISO 3166-1 alpha-3"] = df["ISO 3166-1 alpha-3"].astype("string")

        file_path_sources = pooch.retrieve(
            path=pooch.os_cache("openclimatedata/gcb-fossil"),
            fname=self.filename_sources,
            url=self.url_sources,
            known_hash=self.hash_sources,
        )
        df_sources = pd.read_csv(file_path_sources, encoding="latin-1")

        if "UN M49" in df_sources.columns:
            df_sources = df_sources.drop("UN M49", axis=1)

        if self.published < "2024-10-23":
            # Saint Kitts and Nevis and St. Kitts-Nevis-Anguilla both use KNA.
            # The code is replaced with NaN and treated as below to be differentiable.
            if len(df[df["ISO 3166-1 alpha-3"] == "KNA"].Country.unique()) > 1:
                df["ISO 3166-1 alpha-3"] = df["ISO 3166-1 alpha-3"].replace(
                    "KNA", np.nan
                )
                df_sources["ISO 3166-1 alpha-3"] = df_sources[
                    "ISO 3166-1 alpha-3"
                ].replace("KNA", np.nan)

        # A few islands and Kuwaiti oil fires have no code, reusing country.
        df["ISO 3166-1 alpha-3"] = df["ISO 3166-1 alpha-3"].fillna(df["Country"])
        value_vars = df.columns[2:]
        df = df.reset_index().melt(
            id_vars=["Year", "ISO 3166-1 alpha-3"],
            value_vars=value_vars,
            var_name="Category",
            value_name="Value",
        )
        df = df.rename(columns={"ISO 3166-1 alpha-3": "Code"})

        df_sources["ISO 3166-1 alpha-3"] = df_sources["ISO 3166-1 alpha-3"].fillna(
            df_sources["Country"]
        )
        value_vars = df_sources.columns[2:]
        df_sources = df_sources.reset_index().melt(
            id_vars=["Year", "ISO 3166-1 alpha-3"],
            value_vars=value_vars,
            var_name="Category",
            value_name="Provenance",
        )
        df_sources = df_sources.rename(columns={"ISO 3166-1 alpha-3": "Code"})

        df["Code"] = df["Code"].astype("category")
        df["Category"] = df["Category"].astype("category")
        df_sources["Provenance"] = df_sources["Provenance"].astype("category")

        df = df.set_index(["Year", "Code", "Category"])
        df_sources = df_sources.set_index(["Year", "Code", "Category"])
        return pd.concat([df, df_sources], axis="columns").reset_index()

    def to_ocd(self):
        """Long DataFrame with all column names lower-cased."""
        df = self.to_long_dataframe()
        df.columns = df.columns.map(lambda x: x.lower())
        df["code"] = df["code"].cat.rename_categories({"KSV": "XKX"})
        return df


GCB_Fossil_Emissions = {
    "2025v15": _GCB_Fossil(
        name="The Global Carbon Project's fossil CO2 emissions dataset",
        doi="10.5281/zenodo.17417124",
        published="2025-10-22",
        filename="GCB2025v15_MtCO2_flat.csv",
        url="https://zenodo.org/records/17417124/files/GCB2025v15_MtCO2_flat.csv?download=1",
        hash="md5:3008e30d913af5926a83d0d0775fb72e",
        filename_sources="GCB2025v15_sources_flat.csv",
        url_sources="https://zenodo.org/records/17417124/files/GCB2025v15_sources_flat.csv?download=1",
        hash_sources="md5:bb4e63819ea7fbc33196981ab78567e6",
        citation="""Andrew, R. M., & Peters, G. P. (2025). The Global Carbon Project's fossil CO2 emissions dataset (2025v15) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.17417124""",
        license="CC BY 4.0",
    ),
    "2024v18": _GCB_Fossil(
        name="The Global Carbon Project's fossil CO2 emissions dataset",
        doi="10.5281/zenodo.14106218",
        published="2024-11-13",
        filename="GCB2024v18_MtCO2_flat.csv",
        url="https://zenodo.org/records/14106218/files/GCB2024v18_MtCO2_flat.csv?download=1",
        hash="md5:70dac1843444b14655bf756c70c1f04a",
        filename_sources="GCB2024v18_sources_flat.csv",
        url_sources="https://zenodo.org/records/14106218/files/GCB2024v18_sources_flat.csv?download=1",
        hash_sources="md5:b73b61013613ed685dfc46dba75440d8",
        citation="""Andrew, R. M., & Peters, G. P. (2024). The Global Carbon Project's fossil CO2 emissions dataset (2024v18) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.14106218""",
        license="CC BY 4.0",
    ),
    "2024v17": _GCB_Fossil(
        name="The Global Carbon Project's fossil CO2 emissions dataset",
        doi="10.5281/zenodo.13981696",
        published="2024-10-23",
        filename="GCB2024v17_MtCO2_flat.csv",
        url="https://zenodo.org/records/13981696/files/GCB2024v17_MtCO2_flat.csv",
        hash="md5:70dac1843444b14655bf756c70c1f04a",
        filename_sources="GCB2024v17_sources_flat.csv",
        url_sources="https://zenodo.org/records/13981696/files/GCB2024v17_sources_flat.csv",
        hash_sources="md5:b73b61013613ed685dfc46dba75440d8",
        citation="""Andrew, R. M., & Peters, G. P. (2024). The Global Carbon Project's fossil CO2 emissions dataset (2024v17) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.13981696""",
        license="CC BY 4.0",
    ),
    "2023v43": _GCB_Fossil(
        name="The Global Carbon Project's fossil CO2 emissions dataset",
        doi="10.5281/zenodo.10562476",
        published="2024-01-24",
        filename="GCB2023v43_MtCO2_flat.csv",
        url="https://zenodo.org/records/10562476/files/GCB2023v43_MtCO2_flat.csv",
        hash="md5:e8cc0ffc5b6a4dbc2c1ae2453dcfb859",
        filename_sources="GCB2023v43_sources_flat.csv",
        url_sources="https://zenodo.org/records/10562476/files/GCB2023v43_sources_flat.csv",
        hash_sources="md5:1e88fa3eb7322628b7c4bf1f5a278d97",
        citation="""Andrew, R. M., & Peters, G. P. (2024). The Global Carbon Project's fossil CO2 emissions dataset (2023v43) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.10562476""",
        license="CC BY 4.0",
    ),
    "2023v36": _GCB_Fossil(
        name="The Global Carbon Project's fossil CO2 emissions dataset",
        doi="10.5281/zenodo.10177738",
        published="2023-11-21",
        filename="GCB2023v36_MtCO2_flat.csv",
        url="https://zenodo.org/records/10177738/files/GCB2023v36_MtCO2_flat.csv",
        hash="md5:5bb46f04063157eff3dcdca66c19c553",
        filename_sources="GCB2023v36_sources_flat.csv",
        url_sources="https://zenodo.org/records/10177738/files/GCB2023v36_sources_flat.csv",
        hash_sources="md5:bec410534e85732df5871a08ea2d1322",
        citation="""Andrew, R. M., & Peters, G. P. (2023). The Global Carbon Project's fossil CO2 emissions dataset (2023v36) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.10177738""",
        license="CC BY 4.0",
    ),
    "2023v28": _GCB_Fossil(
        name="The Global Carbon Project's fossil CO2 emissions dataset",
        doi="10.5281/zenodo.10065794",
        published="2023-11-02",
        filename="GCB2023v28_MtCO2_flat.csv",
        url="https://zenodo.org/records/10065794/files/GCB2023v28_MtCO2_flat.csv",
        hash="md5:23d6f3f0a7e88281d41f6eaf8aa68c37",
        filename_sources="GCB2023v28_sources_flat.csv",
        url_sources="https://zenodo.org/records/10065794/files/GCB2023v28_sources_flat.csv",
        hash_sources="md5:f2a5e14d1562f9be80ab8b59c607b9a2",
        citation="""Andrew, R. M., & Peters, G. P. (2023). The Global Carbon Project's fossil CO2 emissions dataset (2023v28) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.10065794""",
        license="CC BY 4.0",
    ),
    "2022v27": _GCB_Fossil(
        name="The Global Carbon Project's fossil CO2 emissions dataset",
        doi="10.5281/zenodo.7215364",
        published="2022-10-17",
        filename="GCB2022v27_MtCO2_flat.csv",
        url="https://zenodo.org/record/7215364/files/GCB2022v27_MtCO2_flat.csv",
        hash="md5:251ce1c5f07d5d28128fa84df856b2f9",
        filename_sources="GCB2022v27_sources_flat.csv",
        url_sources="https://zenodo.org/record/7215364/files/GCB2022v27_sources_flat.csv",
        hash_sources="md5:7ad24d6ec981b55404b0308ed5158d6e",
        citation="""Andrew, Robbie M., & Peters, Glen P. (2022). The Global Carbon Project's fossil CO2 emissions dataset (2022v27) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.7215364""",
        license="CC BY 4.0",
    ),
    "2021v34": _GCB_Fossil(
        name="The Global Carbon Project's fossil CO2 emissions dataset",
        doi="10.5281/zenodo.5569235",
        published="2021-10-14",
        filename="GCB2021v34_MtCO2_flat.csv",
        url="https://zenodo.org/record/5569235/files/GCB2021v34_MtCO2_flat.csv",
        hash="md5:00d432500752936a2c95f6feeb599a51",
        filename_sources="GCB2021v34_sources_flat.csv",
        url_sources="https://zenodo.org/record/5569235/files/GCB2021v34_sources_flat.csv",
        hash_sources="md5:ae185160899541fc4bef08fa869cbde2",
        citation="""Andrew, Robbie M., & Peters, Glen P. (2021). The Global Carbon Project's fossil CO2 emissions dataset (2021v34) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.5569235""",
        license="CC BY 4.0",
    ),
}



import os

import pytest
from pandas.testing import assert_series_equal
from pytest import approx

from openclimatedata import CEDS

GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"

versions = CEDS.keys()


def test_ceds():
    for version in versions:
        # Test __repr__
        print(CEDS)

        assert CEDS[version].name
        assert CEDS[version].doi
        assert CEDS[version].published

        assert len(CEDS[version].entities) > 0


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_ceds_reshaping():
    for version in versions:
        if int(CEDS[version].published[:4]) < 2024:
            df = CEDS[version]["CO"]["by_sector"].to_dataframe()

            if "iso" in df.columns:  # 2019 version
                code = "iso"
                value = 0.00018696315423434
            elif "2018" in version:
                code = "country"
                value = 0.0001748032595938
            else:
                code = "country"
                value = 0.0288382599890555

            assert df[
                (df[code] == "abw") & (df.sector == "1A1a_Electricity-autoproducer")
            ]["X2014"].values[0] == approx(value)
            assert (
                df[
                    (df[code] == "abw") & (df.sector == "1A1a_Electricity-autoproducer")
                ]["X1750"].values[0]
                == 0
            )

            ocdf = CEDS[version]["CO"]["by_sector"].to_ocd()

            assert_series_equal(
                df[
                    (df[code] == "abw") & (df.sector == "1A1a_Electricity-autoproducer")
                ]["X1750"],
                ocdf[
                    (ocdf.code == "ABW")
                    & (ocdf.category == "1A1a_Electricity-autoproducer")
                    & (ocdf.year == 1750)
                ]["value"],
                check_index=False,
                check_names=False,
            )
            assert_series_equal(
                df[
                    (df[code] == "abw") & (df.sector == "1A1a_Electricity-autoproducer")
                ]["X2014"],
                ocdf[
                    (ocdf.code == "ABW")
                    & (ocdf.category == "1A1a_Electricity-autoproducer")
                    & (ocdf.year == 2014)
                ]["value"],
                check_index=False,
                check_names=False,
            )
        elif int(CEDS[version].published[:4]) < 2025:
            df = CEDS[version]["CO"]["by_sector"].to_dataframe()

            value = 0.0146970051332098
            assert df[
                (df["country"] == "abw")
                & (df.sector == "1A1a_Electricity-autoproducer")
            ]["X2022"].values[0] == approx(value)
            assert (
                df[
                    (df["country"] == "abw")
                    & (df.sector == "1A1a_Electricity-autoproducer")
                ]["X1750"].values[0]
                == 0
            )

            ocdf = CEDS[version]["CO"]["by_sector"].to_ocd()

            assert_series_equal(
                df[
                    (df["country"] == "abw")
                    & (df.sector == "1A1a_Electricity-autoproducer")
                ]["X1750"],
                ocdf[
                    (ocdf.code == "ABW")
                    & (ocdf.category == "1A1a_Electricity-autoproducer")
                    & (ocdf.year == 1750)
                ]["value"],
                check_index=False,
                check_names=False,
            )
            assert_series_equal(
                df[
                    (df["country"] == "abw")
                    & (df.sector == "1A1a_Electricity-autoproducer")
                ]["X2022"],
                ocdf[
                    (ocdf.code == "ABW")
                    & (ocdf.category == "1A1a_Electricity-autoproducer")
                    & (ocdf.year == 2022)
                ]["value"],
                check_index=False,
                check_names=False,
            )
        else:
            df = CEDS[version]["CO"]["by_sector"].to_dataframe()

            value = 0.0156595847126071
            assert df[
                (df["country"] == "abw")
                & (df.sector == "1A1a_Electricity-autoproducer")
            ]["X2022"].values[0] == approx(value)
            assert (
                df[
                    (df["country"] == "abw")
                    & (df.sector == "1A1a_Electricity-autoproducer")
                ]["X1750"].values[0]
                == 0
            )

            ocdf = CEDS[version]["CO"]["by_sector"].to_ocd()

            assert_series_equal(
                df[
                    (df["country"] == "abw")
                    & (df.sector == "1A1a_Electricity-autoproducer")
                ]["X1750"],
                ocdf[
                    (ocdf.code == "ABW")
                    & (ocdf.category == "1A1a_Electricity-autoproducer")
                    & (ocdf.year == 1750)
                ]["value"],
                check_index=False,
                check_names=False,
            )
            assert_series_equal(
                df[
                    (df["country"] == "abw")
                    & (df.sector == "1A1a_Electricity-autoproducer")
                ]["X2022"],
                ocdf[
                    (ocdf.code == "ABW")
                    & (ocdf.category == "1A1a_Electricity-autoproducer")
                    & (ocdf.year == 2022)
                ]["value"],
                check_index=False,
                check_names=False,
            )



from ._core import _PRIMAPHIST_2_Release

paper_2016 = """Gütschow, J., Jeffery, L., Gieseke, R., Gebel, R., Stevens, D., Krapp, M., & Rocha, M. (2016): The PRIMAP-hist national historical emissions time series, Earth Syst. Sci. Data, 8, 571-603, doi:10.5194/essd-8-571-2016"""
doi_article = "10.5194/essd-8-571-2016"


citation_2_6_1 = """Gütschow, J., Busch, D., & Pflüger, M. (2025). The PRIMAP-hist national historical emissions time series (1750-2023) v2.6.1 (v2.6.1) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.15016289"""

_PRIMAPHIST_2_6_1 = _PRIMAPHIST_2_Release(
    name="PRIMAP-hist 2.6.1",
    version="2.6.1",
    published="2025-03-13",
    doi="10.5281/zenodo.15016289",
    citation=citation_2_6_1,
    doi_article=doi_article,
    citation_article=paper_2016,
    license="CC BY 4.0",
    files={
        "main": {
            "filename": "Guetschow_et_al_2025-PRIMAP-hist_v2.6.1_final_13-Mar-2025.csv",
            "known_hash": "md5:09b9c61629f87e16012222e5b303bc36",
            "note": "The main dataset with numerical extrapolation of all time series to 2023 and three significant digits.",
        }
    },
)

citation_2_6 = """Gütschow, J., Busch, D., & Pflüger, M. (2024). The PRIMAP-hist national historical emissions time series v2.6 (1750-2023) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.13752654"""

_PRIMAPHIST_2_6 = _PRIMAPHIST_2_Release(
    name="PRIMAP-hist 2.6",
    version="2.6",
    published="2024-09-13",
    doi="10.5281/zenodo.13752654",
    citation=citation_2_6,
    doi_article=doi_article,
    citation_article=paper_2016,
    license="CC BY 4.0",
    files={
        "main": {
            "filename": "Guetschow_et_al_2024a-PRIMAP-hist_v2.6_final_13-Sep-2024.csv",
            "known_hash": "md5:cb58d8d15e3597b5afeef4f1fd096c17",
            "note": "The main dataset with numerical extrapolation of all time series to 2023 and three significant digits.",
        }
    },
)


citation_2_5_1 = """Gütschow, J., Pflüger, M., & Busch, D. (2024). The PRIMAP-hist national historical emissions time series v2.5.1 (1750-2022) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.10705513"""

_PRIMAPHIST_2_5_1 = _PRIMAPHIST_2_Release(
    name="PRIMAP-hist 2.5.1",
    version="2.5.1",
    published="2024-02-27",
    doi="10.5281/zenodo.10705513",
    citation=citation_2_5_1,
    doi_article=doi_article,
    citation_article=paper_2016,
    license="CC BY 4.0",
    files={
        "main": {
            "filename": "Guetschow_et_al_2024-PRIMAP-hist_v2.5.1_final_27-Feb-2024.csv",
            "known_hash": "md5:7de949e90dfc924c3829aa25d32afe1b",
            "note": "The main dataset with numerical extrapolation of all time series to 2022 and three significant digits.",
        }
    },
)

citation_2_5 = """Gütschow, J., & Pflüger, M. (2023). The PRIMAP-hist national historical emissions time series v2.5 (1750-2022) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.10006301"""

_PRIMAPHIST_2_5 = _PRIMAPHIST_2_Release(
    name="PRIMAP-hist 2.5",
    version="2.5",
    published="2023-10-15",
    doi="10.5281/zenodo.10006301",
    citation=citation_2_5,
    doi_article=doi_article,
    citation_article=paper_2016,
    license="CC BY 4.0",
    files={
        "main": {
            "filename": "Guetschow_et_al_2023b-PRIMAP-hist_v2.5_final_15-Oct-2023.csv",
            "known_hash": "md5:77405ac4eb5e915a7a9bfc25ff84d7c8",
            "note": "The main dataset with numerical extrapolation of all time series to 2022 and three significant digits.",
        }
    },
)

citation_2_4_2 = """Gütschow, J., & Pflüger, M. (2023): The PRIMAP-hist national historical emissions time series v2.4.2 (1750-2021) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.7727475"""

_PRIMAPHIST_2_4_2 = _PRIMAPHIST_2_Release(
    name="PRIMAP-hist 2.4.2",
    version="2.4.2",
    doi="10.5281/zenodo.7727475",
    published="2023-03-15",
    citation=citation_2_4_2,
    doi_article=doi_article,
    citation_article=paper_2016,
    license="CC BY 4.0",
    files={
        "main": {
            "filename": "Guetschow-et-al-2023a-PRIMAP-hist_v2.4.2_final_09-Mar-2023.csv",
            "known_hash": "md5:4cd6ac6cf2cc53c211f7bbf6e951e3f1",
            "note": "The main dataset with numerical extrapolation of all time series to 2021 and three significant digits.",
        }
    },
)

citation_2_4_1 = """Gütschow, J., & Pflüger, M. (2023): The PRIMAP-hist national historical emissions time series v2.4.1 (1750-2021) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.7585420"""

_PRIMAPHIST_2_4_1 = _PRIMAPHIST_2_Release(
    name="PRIMAP-hist 2.4.1",
    version="2.4.1",
    doi="10.5281/zenodo.7585420",
    published="2023-02-20",
    citation=citation_2_4_1,
    doi_article=doi_article,
    citation_article=paper_2016,
    license="CC BY 4.0",
    files={
        "main": {
            "filename": "Guetschow-et-al-2023-PRIMAP-hist_v2.4.1_final_16-Feb-2023.csv",
            "known_hash": "md5:220b45725161685f00ba48032f911b28",
            "note": "The main dataset with numerical extrapolation of all time series to 2021 and three significant digits.",
        }
    },
)


citation_2_4 = """Gütschow, J., & Pflüger, M. (2022): The PRIMAP-hist national historical emissions time series v2.4 (1750-2021) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.7179775"""

_PRIMAPHIST_2_4 = _PRIMAPHIST_2_Release(
    name="PRIMAP-hist 2.4",
    version="2.4",
    doi="10.5281/zenodo.7179775",
    published="2022-10-17",
    citation=citation_2_4,
    doi_article=doi_article,
    citation_article=paper_2016,
    license="CC BY 4.0",
    files={
        "main": {
            "filename": "Guetschow-et-al-2022-PRIMAP-hist_v2.4_11-Oct-2022.csv",
            "known_hash": "md5:502b68fb811fac25af8680e91898c324",
            "note": "The main dataset with numerical extrapolation of all time series to 2021 and three significant digits.",
        }
    },
)

citation_2_3_1 = """Gütschow, J., Günther, A., & Pflüger, M. (2021): The PRIMAP-hist national historical emissions time series v2.3.1 (1850-2019) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.5494497"""

_PRIMAPHIST_2_3_1 = _PRIMAPHIST_2_Release(
    name="PRIMAP-hist 2.3.1",
    version="2.3.1",
    doi="10.5281/zenodo.5494497",
    published="2021-09-22",
    citation=citation_2_3_1,
    doi_article=doi_article,
    citation_article=paper_2016,
    license="CC BY 4.0",
    files={
        "main": {
            "filename": "Guetschow-et-al-2021-PRIMAP-hist_v2.3.1_20-Sep_2021.csv",
            "known_hash": "md5:f4cb55a55e4d5e5dcfb1513b677ae318",
            "note": "The main dataset with numerical extrapolation of all time series to 2019 and three significant digits.",
        },
        "no_extrap": {
            "filename": "Guetschow-et-al-2021-PRIMAP-hist_v2.3.1_no_extrap_20-Sep_2021.csv",
            "known_hash": "md5:870bf6d47e74f9245d9f9803d7be80ea",
            "note": "Variant without numerical extrapolation of missing values and not including country groups (three significant digits).",
        },
        "no_extrap_no_rounding": {
            "filename": "Guetschow-et-al-2021-PRIMAP-hist_v2.3.1_no_extrap_no_rounding_20-Sep_2021.csv",
            "known_hash": "md5:f0b9afb73aefbf40693d475ea3dcc6ad",
            "note": "Variant without numerical extrapolation of missing values and not including country groups (eleven significant digits).",
        },
    },
)

citation_2_3 = "Gütschow, J., Günther, A., & Pflüger, M. (2021): The PRIMAP-hist national historical emissions time series v2.3 (1850-2019) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.5175154"

_PRIMAPHIST_2_3 = _PRIMAPHIST_2_Release(
    name="PRIMAP-hist 2.3",
    version="2.3",
    doi="10.5281/zenodo.5175154",
    published="2021-08-30",
    citation=citation_2_3,
    doi_article=doi_article,
    citation_article=paper_2016,
    license="CC BY 4.0",
    files={
        "main": {
            "filename": "Guetschow-et-al-2021-PRIMAP-hist_v2.3_28_Jul_2021.csv",
            "known_hash": "md5:7b29dfb49ca97dcf594bc639767f1e2a",
            "note": "The main dataset with numerical extrapolation of all time series to 2019 and three significant digits.",
        }
    },
)

citation_2_2 = "Gütschow, J., Günther, A., Jeffery, L., & Gieseke, R. (2021): The PRIMAP-hist national historical emissions time series v2.2 (1850-2018) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.4479172"

_PRIMAPHIST_2_2 = _PRIMAPHIST_2_Release(
    name="PRIMAP-hist 2.2",
    version="2.2",
    doi="10.5281/zenodo.4479172",
    published="2021-02-09",
    citation=citation_2_2,
    doi_article=doi_article,
    citation_article=paper_2016,
    license="CC BY 4.0",
    files={
        "main": {
            "filename": "PRIMAP-hist_v2.2_19-Jan-2021.csv",
            "known_hash": "md5:a3a8c25f7b784fdb85c89fbff29f5fd3",
            "note": "With numerical extrapolation of all time series to 2018.",
        }
    },
)


PRIMAPHIST = {
    "2.6.1": _PRIMAPHIST_2_6_1,
    "2.6": _PRIMAPHIST_2_6,
    "2.5.1": _PRIMAPHIST_2_5_1,
    "2.5": _PRIMAPHIST_2_5,
    "2.4.2": _PRIMAPHIST_2_4_2,
    "2.4.1": _PRIMAPHIST_2_4_1,
    "2.4": _PRIMAPHIST_2_4,
    "2.3.1": _PRIMAPHIST_2_3_1,
    "2.3": _PRIMAPHIST_2_3,
    "2.2": _PRIMAPHIST_2_2,
}



import os

import pandas as pd
import pytest
from pytest import approx

from openclimatedata import Global_Carbon_Budget

GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"


versions = Global_Carbon_Budget.keys()


def test_gcb():
    for version in versions:
        assert Global_Carbon_Budget[version].name
        assert Global_Carbon_Budget[version].doi
        assert Global_Carbon_Budget[version].published


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_gcb_sheet_names():
    for version in versions:
        sheet_names = Global_Carbon_Budget[version].keys()

        assert "Global Carbon Budget" in sheet_names
        assert "Historical Budget" in sheet_names
        assert "Fossil Emissions by Category" in sheet_names
        assert "Land-Use Change Emissions" in sheet_names
        assert "Ocean Sink" in sheet_names
        assert "Terrestrial Sink" in sheet_names
        assert "Cement Carbonation Sink" in sheet_names


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_gcb_dataframes_for_subtables_only():
    # In sheets with subtables, `to_dataframe` etc. should only be available
    # for tables, not on the main sheet.
    for version in versions:
        sheet_names = Global_Carbon_Budget[version].keys()

        for sheet_name in sheet_names:
            if len(Global_Carbon_Budget[version][sheet_name].keys()) > 0:
                assert "to_dataframe" not in dir(
                    Global_Carbon_Budget[version][sheet_name]
                )
                assert "to_long_dataframe" not in dir(
                    Global_Carbon_Budget[version][sheet_name]
                )


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_gcb_2023():
    year = "2023"
    global_carbon_budget = Global_Carbon_Budget[year][
        "Global Carbon Budget"
    ].to_dataframe()
    historical_budget = Global_Carbon_Budget[year]["Historical Budget"].to_dataframe()
    fossil_emissions_by_category = Global_Carbon_Budget[year][
        "Fossil Emissions by Category"
    ].to_dataframe()
    luc_emissions_gcb = Global_Carbon_Budget[year]["Land-Use Change Emissions"][
        "GCB"
    ].to_dataframe()
    luc_emisssions_individual_models = Global_Carbon_Budget[year][
        "Land-Use Change Emissions"
    ]["Individual models (NET) - Does not include peat emissions"].to_dataframe()
    ocean_sink_gcb = Global_Carbon_Budget[year]["Ocean Sink"]["GCB"].to_dataframe()
    ocean_sink_data_based_products = Global_Carbon_Budget[year]["Ocean Sink"][
        "Data-based products"
    ].to_dataframe()
    terrestrial_sink_gcb = Global_Carbon_Budget[year]["Terrestrial Sink"][
        "GCB"
    ].to_dataframe()
    terrestrial_sink_individual_models = Global_Carbon_Budget[year]["Terrestrial Sink"][
        "Individual models"
    ].to_dataframe()
    cement_carbonation_sink = Global_Carbon_Budget[year][
        "Cement Carbonation Sink"
    ].to_dataframe()

    assert global_carbon_budget["fossil emissions excluding carbonation"].loc[
        1959
    ] == approx(2.41666545551985)
    assert global_carbon_budget["budget imbalance"].loc[2022] == approx(
        -0.0921741721108096
    )

    assert historical_budget["fossil emissions excluding carbonation"].loc[
        1750
    ] == approx(0.00253982996724891)
    assert historical_budget["budget imbalance"].loc[2022] == approx(
        -0.0921741721108091
    )

    assert fossil_emissions_by_category["fossil.emissions.excluding.carbonation"].loc[
        1850
    ] == approx(53.6986807874229)
    assert fossil_emissions_by_category["Per.Capita"].loc[2022] == approx(
        1.27134792448779
    )

    assert luc_emissions_gcb.Net.loc[1959] == approx(2.12152666666667)
    assert luc_emisssions_individual_models["Model Spread (sd)"].loc[2022] == approx(
        0.607291045665856
    )

    assert ocean_sink_gcb.GCB.loc[1959] == approx(0.992419800030039)
    assert ocean_sink_data_based_products["sd data-products (excl. Watson.)"].loc[
        2022
    ] == approx(0.31727790009081)

    assert terrestrial_sink_gcb.GCB.loc[1959] == approx(0.430358951653902)
    assert terrestrial_sink_individual_models["Model Spread (sd)"].loc[2022] == approx(
        0.829692166272794
    )

    assert cement_carbonation_sink.GCB.loc[1959] == approx(12.542147)
    assert cement_carbonation_sink.GCB.loc[2022] == approx(217.464615)
    assert pd.isna(cement_carbonation_sink.Huang[2022])


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_gcb_2022():
    year = "2022"
    global_carbon_budget = Global_Carbon_Budget[year][
        "Global Carbon Budget"
    ].to_dataframe()
    historical_budget = Global_Carbon_Budget[year]["Historical Budget"].to_dataframe()
    fossil_emissions_by_category = Global_Carbon_Budget[year][
        "Fossil Emissions by Category"
    ].to_dataframe()
    luc_emissions_gcb = Global_Carbon_Budget[year]["Land-Use Change Emissions"][
        "GCB"
    ].to_dataframe()
    luc_emisssions_individual_models = Global_Carbon_Budget[year][
        "Land-Use Change Emissions"
    ]["Individual models"].to_dataframe()
    ocean_sink_gcb = Global_Carbon_Budget[year]["Ocean Sink"]["GCB"].to_dataframe()
    ocean_sink_data_based_products = Global_Carbon_Budget[year]["Ocean Sink"][
        "Data-based products"
    ].to_dataframe()
    terrestrial_sink_gcb = Global_Carbon_Budget[year]["Terrestrial Sink"][
        "GCB"
    ].to_dataframe()
    terrestrial_sink_individual_models = Global_Carbon_Budget[year]["Terrestrial Sink"][
        "Individual models"
    ].to_dataframe()
    cement_carbonation_sink = Global_Carbon_Budget[year][
        "Cement Carbonation Sink"
    ].to_dataframe()

    assert global_carbon_budget["fossil emissions excluding carbonation"].loc[
        1959
    ] == approx(2.41709093103705)
    assert global_carbon_budget["budget imbalance"].loc[2021] == approx(
        -0.57757145051883
    )

    assert historical_budget["fossil emissions excluding carbonation"].loc[
        1750
    ] == approx(0.002552)
    assert historical_budget["budget imbalance"].loc[2021] == approx(-0.57757145051883)

    assert fossil_emissions_by_category["fossil.emissions.excluding.carbonation"].loc[
        1850
    ] == approx(53.738)
    assert fossil_emissions_by_category["Per.Capita"].loc[2021] == approx(
        1.28103137304945
    )

    assert luc_emissions_gcb.Net.loc[1959] == approx(1.93893333333333)
    assert luc_emisssions_individual_models["Model Spread (sd)"].loc[2021] == approx(
        0.488150134956261
    )

    assert ocean_sink_gcb.GCB.loc[1959] == approx(0.975005218749614)
    assert ocean_sink_data_based_products["sd data-products (excl. Watson.)"].loc[
        2021
    ] == approx(0.24795776316279)

    assert terrestrial_sink_gcb.GCB.loc[1959] == approx(0.401805228689998)
    assert terrestrial_sink_individual_models["Model Spread (sd)"].loc[2021] == approx(
        0.888433352902292
    )

    assert cement_carbonation_sink.GCB.loc[1959] == approx(12.684256)
    assert cement_carbonation_sink.GCB.loc[2021] == approx(229.794956)
    assert pd.isna(cement_carbonation_sink.Guo[2021])


@pytest.mark.skipif(GITHUB_ACTIONS, reason="Test requires downloading.")
def test_gcb_2021():
    year = "2021"
    global_carbon_budget = Global_Carbon_Budget[year][
        "Global Carbon Budget"
    ].to_dataframe()
    historical_budget = Global_Carbon_Budget[year]["Historical Budget"].to_dataframe()
    fossil_emissions_by_category = Global_Carbon_Budget[year][
        "Fossil Emissions by Category"
    ].to_dataframe()
    luc_emissions_gcb = Global_Carbon_Budget[year]["Land-Use Change Emissions"][
        "GCB"
    ].to_dataframe()
    luc_emisssions_individual_models = Global_Carbon_Budget[year][
        "Land-Use Change Emissions"
    ]["Individual models"].to_dataframe()
    ocean_sink_gcb = Global_Carbon_Budget[year]["Ocean Sink"]["GCB"].to_dataframe()
    ocean_sink_data_based_products = Global_Carbon_Budget[year]["Ocean Sink"][
        "Data-based products"
    ].to_dataframe()
    terrestrial_sink_gcb = Global_Carbon_Budget[year]["Terrestrial Sink"][
        "GCB"
    ].to_dataframe()
    terrestrial_sink_individual_models = Global_Carbon_Budget[year]["Terrestrial Sink"][
        "Individual models"
    ].to_dataframe()
    cement_carbonation_sink = Global_Carbon_Budget[year][
        "Cement Carbonation Sink"
    ].to_dataframe()

    assert global_carbon_budget["fossil emissions excluding carbonation"].loc[
        1959
    ] == approx(2.41713282396274)
    assert global_carbon_budget["budget imbalance"].loc[2020] == approx(
        -0.775025676533141
    )

    assert historical_budget["fossil emissions excluding carbonation"].loc[
        1750
    ] == approx(0.002552)
    assert historical_budget["budget imbalance"].loc[2020] == approx(-0.775025676531002)

    assert fossil_emissions_by_category["fossil.emissions.excluding.carbonation"].loc[
        1959
    ] == approx(2417.13282396274)
    assert fossil_emissions_by_category["Per.Capita"].loc[2020] == approx(
        1.21875600594367
    )

    assert luc_emissions_gcb.Net.loc[1959] == approx(1.84551880806033)
    assert luc_emisssions_individual_models["Model Spread (sd)"].loc[2020] == approx(
        0.730856668977054
    )

    assert ocean_sink_gcb.GCB.loc[1959] == approx(0.918022208427885)
    assert ocean_sink_data_based_products["sd data-products (excl. Watson.)"].loc[
        2020
    ] == approx(0.33545744617825)

    assert terrestrial_sink_gcb.GCB.loc[1959] == approx(0.395882352941176)
    assert terrestrial_sink_individual_models["Model Spread (sd)"].loc[2020] == approx(
        0.989381862876127
    )

    assert cement_carbonation_sink.GCB.loc[1959] == approx(12.68373)
    assert cement_carbonation_sink.GCB.loc[2020] == approx(218.796424)
    assert pd.isna(cement_carbonation_sink.Guo[2020])



import re
from dataclasses import dataclass
from typing import Optional

import pandas as pd
import pooch
from openpyxl import load_workbook


class _Global_Carbon_Budget_Release(dict):
    def __init__(
        self,
        name: str,
        version: str,
        doi: str,
        doi_article: str,
        published: str,
        citation: str,
        citation_article: str,
        license: str,
        filename: str,
        url: str,
        known_hash: str,
        sheets: list,
    ):
        self.name = name
        self.version = version
        self.doi = doi
        self.doi_article = doi_article
        self.published = published
        self.citation = citation
        self.citation_article = citation_article
        self.license = license
        self.filename = filename
        self.url = url
        self.known_hash = known_hash

        for sheet in sheets:
            self[sheet["sheet_name"]] = _Global_Carbon_Budget_Sheet(
                release=self, **sheet
            )

    def _get_file_path(self):
        return pooch.retrieve(
            path=pooch.os_cache("openclimatedata/global-carbon-budget"),
            fname=self.filename,
            url=self.url,
            known_hash=self.known_hash,
        )

    def __repr__(self):
        return f"""{self.name}

'{self.filename}'

License: {self.license}
https://doi.org/{self.doi}

{self.citation}"""


class _Global_Carbon_Budget_Sheet(dict):
    def __init__(
        self,
        release: object,
        sheet_name: str,
        skiprows: int,
        columns: Optional[str] = None,
        tables: Optional[list] = None,
    ):
        self.release = release
        self.sheet_name = sheet_name
        self.skiprows = skiprows
        self.columns = columns

        if tables:
            for table in tables:
                self[table["table_name"]] = _Global_Carbon_Budget_Table(
                    sheet=self,
                    table_name=table["table_name"],
                    skiprows=table["skiprows"],
                    columns=table["columns"],
                )
        else:
            self.to_dataframe = self._to_dataframe
            self.to_long_dataframe = self._to_long_dataframe

    def __repr__(self):
        if not hasattr(self, "note"):
            file_path = self.release._get_file_path()
            wb = load_workbook(file_path)
            rows = [
                list(row)
                for row in list(
                    wb[self.sheet_name].iter_rows(
                        max_row=self.skiprows, values_only=True
                    )
                )
            ]
            note = ""
            for row in rows:
                note += ("\t".join([cell if cell else "\t" for cell in row])).strip()
                note += "\n"

            self.note = note

        return f"""{self.release.name} — {self.sheet_name}

{self.note}
"""

    def _to_dataframe(self):
        file_path = self.release._get_file_path()
        df = pd.read_excel(
            file_path,
            sheet_name=self.sheet_name,
            skiprows=self.skiprows,
            usecols=self.columns,
            index_col=0,
        )
        df.index.name = "Year"
        return df

    def _to_long_dataframe(self):
        df = self.to_dataframe()
        value_vars = df.columns
        df = df.reset_index().melt(
            id_vars=["Year"],
            value_vars=value_vars,
            var_name="Category",
            value_name="Value",
        )
        df.Category = df.Category.astype("category")
        return df


@dataclass
class _Global_Carbon_Budget_Table:
    sheet: object
    table_name: str
    skiprows: int
    columns: str

    def __repr__(self):
        return f"""{self.sheet.sheet_name} - {self.table_name} - {self.columns}"""

    def to_dataframe(self):
        file_path = self.sheet.release._get_file_path()
        df = pd.read_excel(
            file_path,
            sheet_name=self.sheet.sheet_name,
            skiprows=self.skiprows,
            usecols=self.columns,
            index_col=0,
        )
        # Remove suffixes from duplicated columns names (added by Pandas).
        df.columns = df.columns.map(lambda x: re.sub(r"\.\d$", "", x))
        df.name = self.table_name
        df.index.name = "Year"
        return df

    def to_long_dataframe(self):
        df = self.to_dataframe()
        value_vars = df.columns
        df = df.reset_index().melt(
            id_vars=["Year"],
            value_vars=value_vars,
            var_name="Category",
            value_name="Value",
        )
        df.Category = df.Category.astype("category")
        return df



from ._core import _Global_Carbon_Budget_Release

GCB2023 = _Global_Carbon_Budget_Release(
    name="Global Carbon Budget 2023",
    version="1.1",
    doi="10.18160/GCP-2023",
    doi_article="10.5194/essd-15-5301-2023",
    published="2023-12-05",
    citation="Global Carbon Project, 2023. Supplemental data of Global Carbon Budget 2023. https://doi.org/10.18160/GCP-2023",
    citation_article="Friedlingstein, P., O'Sullivan, M., Jones, M. W., Andrew, R. M., Bakker, D. C. E., Hauck, J., Landschützer, P., Le Quéré, C., Luijkx, I. T., Peters, G. P., Peters, W., Pongratz, J., Schwingshackl, C., Sitch, S., Canadell, J. G., Ciais, P., Jackson, R. B., Alin, S. R., Anthoni, P., Barbero, L., Bates, N. R., Becker, M., Bellouin, N., Decharme, B., Bopp, L., Brasika, I. B. M., Cadule, P., Chamberlain, M. A., Chandra, N., Chau, T.-T.-T., Chevallier, F., Chini, L. P., Cronin, M., Dou, X., Enyo, K., Evans, W., Falk, S., Feely, R. A., Feng, L., Ford, D. J., Gasser, T., Ghattas, J., Gkritzalis, T., Grassi, G., Gregor, L., Gruber, N., Gürses, Ö., Harris, I., Hefner, M., Heinke, J., Houghton, R. A., Hurtt, G. C., Iida, Y., Ilyina, T., Jacobson, A. R., Jain, A., Jarníková, T., Jersild, A., Jiang, F., Jin, Z., Joos, F., Kato, E., Keeling, R. F., Kennedy, D., Klein Goldewijk, K., Knauer, J., Korsbakken, J. I., Körtzinger, A., Lan, X., Lefèvre, N., Li, H., Liu, J., Liu, Z., Ma, L., Marland, G., Mayot, N., McGuire, P. C., McKinley, G. A., Meyer, G., Morgan, E. J., Munro, D. R., Nakaoka, S.-I., Niwa, Y., O'Brien, K. M., Olsen, A., Omar, A. M., Ono, T., Paulsen, M., Pierrot, D., Pocock, K., Poulter, B., Powis, C. M., Rehder, G., Resplandy, L., Robertson, E., Rödenbeck, C., Rosan, T. M., Schwinger, J., Séférian, R., Smallman, T. L., Smith, S. M., Sospedra-Alfonso, R., Sun, Q., Sutton, A. J., Sweeney, C., Takao, S., Tans, P. P., Tian, H., Tilbrook, B., Tsujino, H., Tubiello, F., van der Werf, G. R., van Ooijen, E., Wanninkhof, R., Watanabe, M., Wimart-Rousseau, C., Yang, D., Yang, X., Yuan, W., Yue, X., Zaehle, S., Zeng, J., and Zheng, B.: Global Carbon Budget 2023, Earth Syst. Sci. Data, 15, 5301–5369, https://doi.org/10.5194/essd-15-5301-2023, 2023.",
    license="CC BY 4.0",
    filename="Global_Carbon_Budget_2023v1.1.xlsx",
    url="https://data.icos-cp.eu/licence_accept?ids=%5B%22NMvAsIKjrLx4KUeha_ckfVPP%22%5D",
    known_hash="34cbc0b082a3acbc782947a16bf7247d53cf867ab4a0cd241e8491f2d00842b9",
    sheets=[
        {
            "sheet_name": "Global Carbon Budget",
            "skiprows": 21,
        },
        {
            "sheet_name": "Historical Budget",
            "skiprows": 15,
        },
        {"sheet_name": "Fossil Emissions by Category", "skiprows": 8},
        {
            "sheet_name": "Land-Use Change Emissions",
            "skiprows": 34,
            "tables": [
                {
                    "table_name": "GCB",
                    "skiprows": 37,
                    "columns": "A:G",
                },
                {
                    "table_name": "BLUE",
                    "skiprows": 37,
                    "columns": "A,I:M",
                },
                {
                    "table_name": "H&C2023",
                    "skiprows": 37,
                    "columns": "A,N:R",
                },
                {
                    "table_name": "OSCAR",
                    "skiprows": 37,
                    "columns": "A,S:W",
                },
                {
                    "table_name": "Peat Drainage & Peat Fires",
                    "skiprows": list(range(36)) + [37],
                    "columns": "A,Y:AA",
                },
                {
                    "table_name": "Individual models (NET) - Does not include peat emissions",
                    "skiprows": list(range(36)) + [37],
                    "columns": "A,AC:AV,AX:AY",
                },
            ],
        },
        {
            "sheet_name": "Ocean Sink",
            "skiprows": 28,
            "tables": [
                {"table_name": "GCB", "skiprows": 30, "columns": "A:C"},
                {
                    "table_name": "Individual models",
                    "skiprows": 30,
                    "columns": "A,E:N,P:Q",
                },
                {
                    "table_name": "Data-based products",
                    "skiprows": 30,
                    "columns": "A,S:AB",
                },
            ],
        },
        {
            "sheet_name": "Terrestrial Sink",
            "skiprows": 26,
            "tables": [
                {"table_name": "GCB", "skiprows": 27, "columns": "A:B"},
                {
                    "table_name": "Individual models",
                    "skiprows": 27,
                    "columns": "A,D:W,Y:Z",
                },
            ],
        },
        {"sheet_name": "Cement Carbonation Sink", "skiprows": 9, "columns": "A,B,D,E"},
    ],
)



from dataclasses import dataclass

import pandas as pd
import pooch


@dataclass
class _GCB_National:
    sheet_name: str
    skiprows: int
    note: str
    methods: str
    citation: str
    name: str
    doi: str
    filename: str
    url: str
    hash: str
    license: str

    def __repr__(self):
        return f"""{self.name}
'{self.filename}' - '{self.sheet_name}'

License: {self.license}
https://doi.org/{self.doi}

{self.note}

{self.citation}"""

    def to_dataframe(self):
        file_path = pooch.retrieve(
            path=pooch.os_cache("openclimatedata"),
            fname=self.filename,
            url=self.url,
            known_hash=self.hash,
        )
        return pd.read_excel(
            file_path, sheet_name=self.sheet_name, skiprows=self.skiprows, index_col=0
        )

    def to_long_dataframe(self):
        df = self.to_dataframe()
        df.index.name = "Year"
        value_vars = df.columns
        return df.reset_index().melt(
            id_vars=["Year"],
            value_vars=value_vars,
            var_name="Area",
            value_name="Value",
        )


GCB_National_Emissions = {
    "2022_territorial": _GCB_National(
        name="National Fossil Carbon Emissions 2022 v1.0",
        doi="10.18160/gcp-2022",
        filename="National_Fossil_Carbon_Emissions_2022v1.0.xlsx",
        url="https://data.icos-cp.eu/licence_accept?ids=%5B%22zL1wtJrG7Q5xdvF39Ylg3lUw%22%5D",
        hash="sha256:ccbd70b49ac6ed0e7176f177f58960de5530034623241c6ca7d5b33ae6af5c01",
        sheet_name="Territorial Emissions",
        skiprows=11,
        note="""Fossil CO2 emissions by country (territorial)
All values in million tonnes of carbon per year. For values in million tonnes of CO2 per year, multiply the values below by 3.664
1MtC = 1 million tonne of carbon = 3.664 million tonnes of CO2
""",
        citation="""Cite as: Friedlingstein et al. (2022)""",
        methods="""Full details of the method are described in Friedlingstein et al (2022) and Andrew and Peters (2021)
(1) National estimates include emissions from fossil fuel combustion and oxidation and cement production and excludes emissions from bunker fuels. World totals include emissions from bunker fuels.
(2) Bunker fuels: Emissions from fuels used for international aviation and maritime transport
(3) The disaggregations of regions (e.g. the former Soviet Union prior to 1992) are based on the shares of emissions in the first year after the countries are disaggregated (e.g., 1992 for the Former Soviet Union).
(4) The statistical difference presented on column HX is the difference between the world emissions and the sum of the emissions for each countries and for the bunker fuels.
""",
        license="CC BY 4.0",
    ),
    "2021_territorial": _GCB_National(
        name="National Fossil Carbon Emissions 2021 v1.0",
        doi="10.18160/gcp-2021",
        filename="National_Carbon_Emissions_2021v1.0.xlsx",
        url="https://data.icos-cp.eu/licence_accept?ids=%5B%22lApekzcmd4DRC34oGXQqOxbJ%22%5D",
        hash="sha256:940a5e9337267780d10b7e2819742a3b16c97fa41132274e4c5441696c991a70",
        sheet_name="Territorial Emissions",
        skiprows=11,
        note="""Fossil CO2 emissions by country (territorial)
All values in million tonnes of carbon per year. For values in million tonnes of CO2 per year, multiply the values below by 3.664
1MtC = 1 million tonne of carbon = 3.664 million tonnes of CO2""",
        citation="""Cite as: Friedlingstein et al. 2021""",
        methods="""Methods: Full details of the method are described in Friedlingstein et al. (2021) and Andrew and Peters (2021)
(1) National estimates include emissions from fossil fuel combustion and oxidation and cement production and excludes emissions from bunker fuels. World totals include emissions from bunker fuels.
(2) Bunker fuels: Emissions from fuels used for international aviation and maritime transport
(3) The disaggregations of regions (e.g. the former Soviet Union prior to 1992) are based on the shares of emissions in the first year after the countries are disaggregated (e.g., 1992 for the Former Soviet Union).
(4) The statistical difference presented on column HX is the difference between the world emissions and the sum of the emissions for each countries and for the bunker fuels.""",
        license="CC BY 4.0",
    ),
}
